import logging
import struct
from ..exceptions import NoBindingException
from ..tools import is_fqdn
from .activation_blob import ActivationBlob
from .interface import NdrInterface
from .objref_custom import ObjRefCustom
from .orpcthat import ORPCTHAT
from .props_out_info import PropsOutInfo
from .scm_reply_info_data import ScmReplyInfoData


class RemoteCreateInstanceResponse(NdrInterface):

    FMT1_32 = '<LLL'
    FMT1_64 = '<QLL'

    FMT1_32_SZ = struct.calcsize(FMT1_32)

    def __init__(self, target: str, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)

        # activation_blobs
        (
            self.referent_id,
            _,
            size
        ) = struct.unpack_from(self.FMT1_32, data, offset)
        offset += self.FMT1_32_SZ

        self.objref = ObjRefCustom.from_data(data, offset, size)
        offset += size

        ab_data = ActivationBlob.from_data(self.objref.object_data)

        self.error_code, = struct.unpack_from('<L', data, offset)
        self.props_out_info = PropsOutInfo(ab_data.properties[0])
        self.scm_reply_info_data = ScmReplyInfoData(ab_data.properties[1])
        self._binding = None
        target = target.upper()
        self._target = target.partition('.')[0] if is_fqdn(target) else target

        assert self.error_code == 0, f'error code: {self.error_code}'

    def get_binding(self):
        if self._binding is None:
            for bindingtuple in self.scm_reply_info_data.str_bindings:
                tower_id, binding = bindingtuple
                if tower_id != 7:
                    continue

                if binding.find('[') >= 0:
                    binding, _, port = binding.strip(']\x00').partition('[')
                    port = int(port)
                else:
                    port = 0

                self._binding = (binding, port)

                if binding.upper().find(self._target) >= 0:
                    # take this binding, otherwise just the last one
                    break

        if self._binding is None:
            raise NoBindingException('no network binding has been found')

        logging.debug(f'selected binding: {self._binding}')
        return self._binding

    def get_ipid(self) -> int:
        return self.props_out_info.objref.ipid



import struct
from .orpcthat import ORPCTHAT
from .objref_standard import ObjRefStandard
from .interface import NdrInterface
from .objref import ObjRef
from ..uuid import uuid_to_bin
from uuid import uuid4


class GetSmartEnumResponse(NdrInterface):

    FMT1_32 = '<LLL'
    FMT1_64 = '<QLL'

    FMT1_32_SZ = struct.calcsize(FMT1_32)

    def __init__(self, data: bytes):
        self.orpcthat, offset = ORPCTHAT.from_data(data, offset=0)

        # activation_blobs
        (
            self.referent_id,
            ul_cnt_data_mi,
            ul_cnt_data_ma,
        ) = struct.unpack_from(self.FMT1_32, data, offset=offset)
        offset += self.FMT1_32_SZ

        self.objref: ObjRefStandard =\
            ObjRef.from_data(data, offset, ul_cnt_data_ma)

        offset += ul_cnt_data_ma
        self.error_code, = struct.unpack_from('<L', data, offset)

        # generate a random proxy guid
        self.proxy_guid = uuid_to_bin(str(uuid4()))

        assert self.error_code == 0, f'error code: {self.error_code}'

    def get_ipid(self):
        return self.objref.ipid



import struct
from .common import RpcCommon
from .cont_elem import RpcContElem
from .const import MSRPC_BIND


# 12.6.4.5 The bind_nak PDU
# The IDL declaration of the bind_nak PDU is as follows:
# typedef struct {
#   /* include common fields */
#   u_int16 max_xmit_frag; /* 16:02 max transmit frag size, bytes */
#   u_int16 max_recv_frag; /* 18:02 max receive frag size, bytes */
#   u_int32 assoc_group_id; /* 20:04 incarnation of client-server
#   struct {
#     u_int8 n_context_elem; /* number of items */
#     u_int8 reserved; /* alignment pad, m.b.z. */
#     u_short reserved2; /* alignment pad, m.b.z. */
#     p_cont_elem_t [size_is(n_cont_elem)] p_cont_elem[];
#   }
#   /* optional authentication verifier */
#   /* following fields present iff auth_length != 0 */
#   auth_verifier_co_t auth_verifier;
# }


class RpcBind(RpcCommon):
    FMT = '<HHLBBH'

    def __init__(self):
        super().init(MSRPC_BIND)

        self._max_xmit_frag = 0x10b8
        self._max_recv_frag = 0x10b8
        self._assoc_group_id = 0
        self._n_context_elem = 0
        self._reserved = 0
        self._reserved2 = 0
        self._p_cont_elem = []

    def add_cont_elem(self, cont_elem: RpcContElem) -> None:
        """Add Context Element."""
        cont_elem.p_cont_id = self._n_context_elem
        self._p_cont_elem.append(cont_elem)
        self._n_context_elem += 1

    def freeze_context(self) -> int:
        """Freeze the context and return the context length."""
        pdu_data = struct.pack(
            self.FMT,
            self._max_xmit_frag,            # 1. <H
            self._max_recv_frag,            # 2. <H
            self._assoc_group_id,           # 3. <L
            self._n_context_elem,           # 3. B
            self._reserved,                 # 4. B
            self._reserved2,                # 5. <H
        ) + b''.join(cont_elem.get_data() for cont_elem in self._p_cont_elem)
        return self.set_pdu_data(pdu_data)



import struct
from .common import NdrCommon
from ..uuid import bin_to_str, CLSID_SZ
from .objref import ObjRef
from .objref_standard import ObjRefStandard


class PropsOutInfo(NdrCommon):

    FMT1_32 = '<LLLLL'
    FMT1_32_SZ = struct.calcsize(FMT1_32)

    FMT1_64 = '<LQQQL'
    FMT1_64_SZ = struct.calcsize(FMT1_64)

    FMT2_32 = '<LLLLLL'
    FMT2_32_SZ = struct.calcsize(FMT2_32)

    FMT2_64 = '<LLLQLL'
    FMT2_64_SZ = struct.calcsize(FMT2_64)

    def __init__(self, data: bytes):
        offset = 0
        self.read_common(data, offset)
        offset += self.COMMON_SZ
        self.read_private(data, offset)
        offset += self.PRIVATE_SZ

        (
            c_ifs,
            referent_id1,
            referent_id2,
            referent_id3,
            c_ifs,
        ) = struct.unpack_from(self.FMT1_32, data, offset=offset)
        offset += self.FMT1_32_SZ

        self.piids = []
        for _ in range(c_ifs):
            piid = bin_to_str(data, offset)
            offset += CLSID_SZ
            self.piids.append(piid)

        # TODO: the is a HRESULT
        (
            n_result,
            hr_result,
            n_result,
            referent_id4,
            ul_cnt_data_mi,
            ul_cnt_data_ma,
        ) = struct.unpack_from(self.FMT2_32, data, offset=offset)
        offset += self.FMT2_32_SZ

        self.objref: ObjRefStandard =\
            ObjRef.from_data(data, offset, ul_cnt_data_ma)



import logging
import asyncio
import struct
from typing import Optional, Callable
from .dcom import Dcom
from .rpc.common import RpcCommon
from .rpc.response import RpcResponse
from .rpc.request import RpcRequest
from .rpc.fault import RpcFault
from .ndr.interface import NdrInterface
from .rpc.const import PFC_LAST_FRAG
from .rpc.baseresp import RpcBaseResp
from .request import Request
from .exceptions import DcomException
from .buf import Buf


class Protocol(asyncio.Protocol):

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        self._transport = None
        self._buf = None
        self._tmp = b''
        self._requests = {}
        self._dcom = Dcom()

        # The properties below will be set by connection.py
        self._interface: NdrInterface = None
        self._auth_type: int = None
        self._auth_level: int = None
        self._flags = None
        self._client_seal: Optional[Callable] = None
        self._server_seal: Optional[Callable] = None
        self._iid = None

    @staticmethod
    def get_call_id(data: bytes):
        call_id, = struct.unpack_from('<L', data, offset=12)
        return call_id

    def connection_made(self, transport: asyncio.Transport) -> None:
        '''
        override asyncio.Protocol
        '''

        self._transport = transport
        logging.info(f'connection made: {self.connection_info()}')

    def connection_lost(self, exc: Exception) -> None:
        '''
        override asyncio.Protocol
        '''
        logging.info(f'connection lost {self.connection_info()}')
        self._transport = None

    def __bool__(self):
        return self._transport is not None

    def data_received(self, data: bytes) -> None:
        """override asyncio.Protocol

        Keep a global buffer for fragmented data on the socket, and separate
        buffers for each request since parts of the requests may be received
        within diffrent package fragments.
        """
        # print('RECV!!')
        # print(data)
        if self._buf is None:
            data = self._tmp + data

            if len(data) < RpcCommon.COMMON_SIZE:
                # We do not have a complete header, thus no call_id and size.
                # Use a temporary buffer until we have a complete header.
                self._tmp = data
                return

            size, _, call_id, = struct.unpack_from('<HHL', data, offset=8)
            self._buf = Buf(size, call_id)
            self._tmp = b''

        more = self._buf.append(data)

        if more is False:
            return None

        req, data, = self._requests[self._buf.call_id], self._buf.data

        self._buf = None

        if req.size is not None:
            assert req.fut  # when size is set, we must have a future
            req.buf += data

            if len(req.buf) < req.size:
                if more:
                    self.data_received(more)
                return

            data = req.buf[:req.size]
            rest = req.buf[req.size:]
            req.size = None
            req.buf = rest

        if req.fut:
            req.fut.set_result(data)
            req.fut = None
        elif data:
            req.buf += data

        if more:
            self.data_received(more)

    def write(self, data: bytes):
        call_id, = struct.unpack_from('<L', data, offset=12)
        # print('SEND!! Call Id:', call_id)
        # print(data)
        self._transport.write(data)

    def connection_info(self) -> str:
        if self._transport is None:
            return 'disconnected'
        socket = self._transport.get_extra_info('socket', None)
        if socket is None:
            return 'unknown_addr'
        addr, port = socket.getpeername()[:2]
        return f'{addr}:{port}'

    def close(self):
        # close open requests, if any
        for req in self._requests.values():
            if req.fut is not None:
                req.fut.cancel()
        self._requests.clear()

        # Clear interface
        self._interface = None

        # close transport, if exists
        if self._transport is None:
            return
        self._transport.close()

    async def get_dcom_response(
            self,
            request: bytes,
            size: Optional[int] = None) -> RpcBaseResp:
        req = Request(size=size)
        call_id = self.get_call_id(request)

        assert call_id not in self._requests
        self._requests[call_id] = req
        try:
            pdu_data_list = []
            self.write(request)

            data = await req.fut

            while True:
                rpc_common = RpcCommon.from_data(data)
                ndata = len(data)

                callback = self._dcom._DCOM_RPC_MAP.get(rpc_common.ptype)
                if callback is None:
                    raise DcomException(f'Unknown ptype: {rpc_common.ptype}')
                response = callback(self._dcom, rpc_common, data)

                n = response.rpc_common.frag_length
                if n > ndata:
                    n -= RpcResponse.SIZE
                    resp = await req.readn(n)
                    pdu_data_list.append((resp, n))

                if rpc_common.pfc_flags & PFC_LAST_FRAG:
                    break

                data = await req.readn(size)

            if pdu_data_list:
                response.set_pdu_data_list(pdu_data_list)

            if isinstance(response, RpcFault):
                response.throw()

        finally:
            self._requests.pop(call_id)

        return response



"""ORPCTHIS

See MS-DCOM, 2.2.13.3
"""
import struct
from ..tools import gen_cid
from .com_version import COM_VERSION


class ORPCTHIS:

    EXT32 = '<L'
    EXT64 = '<Q'

    @classmethod
    def get_data(cls, flags: int = 0) -> bytes:
        flags = flags
        reserved = 0
        cid = gen_cid()
        extensions = 0

        return COM_VERSION + struct.pack(
            '<LL',
            flags,
            reserved
        ) + cid + struct.pack(cls.EXT32, extensions)



import struct

# struct {
#   p_context_id_t p_cont_id; (u_int16)
#   u_int8 n_transfer_syn; /* number of items */
#   u_int8 reserved; /* alignment pad, m.b.z. */
#   p_syntax_id_t abstract_syntax; /* transfer syntax list */
#   p_syntax_id_t [size_is(n_transfer_syn)] transfer_syntaxes[];
# } p_cont_elem_t


class RpcContElem:
    FMT = '<HBB'

    def __init__(self, abstract_syntax):
        # this is the context Id, the index number in the parent list
        self.p_cont_id = None

        self._n_transfer_syn = 0
        self._reserved = 0
        self._abstract_syntax = abstract_syntax
        self._transfer_syntaxes = []

    def add_transfer_syntax(self, transfer_syntax):
        self._transfer_syntaxes.append(transfer_syntax)
        self._n_transfer_syn += 1

    def get_data(self):
        return struct.pack(
            self.FMT,
            self.p_cont_id,
            self._n_transfer_syn,
            self._reserved
        ) + self._abstract_syntax + b''.join(self._transfer_syntaxes)



import logging
import asyncio
from typing import Optional, Callable
from Crypto.Cipher import ARC4
from .protocol import Protocol
from .dcom import Dcom
from .rpc.bind_ack import RpcBindAck
from .ntlm.auth_challange import NTLMAuthChallenge
from .ntlm.auth_authenticate import NTLMAuthAuthenticate
from .ntlm.auth_negotiate import NTLMAuthNegotiate
from .ntlm.tools import seal_func, sign_func
from .tools import get_rangom_bytes, encrypted_session_key
from .rpc.const import (
    RPC_C_AUTHN_LEVEL_CONNECT,
    RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,
    RPC_C_AUTHN_LEVEL_PKT_PRIVACY,
    RPC_C_AUTHN_WINNT
)
from .ntlm.const import (
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    NTLMSSP_NEGOTIATE_KEY_EXCH,
)
from .ntlm.tools import (
    seal_func,
    seal_key,
    sign_func,
    sign_key,
)
from .ndr.remote_create_instance import RemoteCreateInstance
from .dcom_const import (
    CLSID_IWbemLevel1Login,
    IID_IRemoteSCMActivator,
    IID_IWbemLevel1Login,
    IID_IWbemServices,
)
from .rpc.request import RpcRequest
from .tools import pad4
from .rpc.response import RpcResponse
from .rpc.auth_verifier_co import RpcAuthVerifierCo
from .ndr.remote_create_instance_response import RemoteCreateInstanceResponse
from .ndr.ntlm_login_response import NTLMLoginResponse
from .ntlm.login import NTLMLogin


class Connection:

    def __init__(
            self,
            host: str,
            username: str,
            password: str,
            domain: str = '',
            port: int = 135,
            loop: Optional[asyncio.AbstractEventLoop] = None):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._domain = domain
        self._loop = asyncio.get_event_loop() if loop is None else loop
        self._protocol: Optional[Protocol] = None
        self._timeout: int = 5
        self._namespace: Optional[str] = None

    async def connect(self, timeout: int = 5):
        conn = self._loop.create_connection(
            lambda: Protocol(loop=self._loop),
            host=self._host,
            port=self._port)

        _, self._protocol = await asyncio.wait_for(
            conn,
            timeout=timeout)
        self._timeout = timeout

    def close(self):
        if self._protocol:
            self._protocol.close()
        self._protocol = None

    def is_connected(self) -> bool:
        return bool(self._protocol)

    def connection_info(self) -> str:
        if not self.is_connected():
            return 'disconnected'
        return self._protocol.connection_info()

    async def _bind(self, iid: bytes, proto: Protocol):
        ntlm_auth_negotiate = NTLMAuthNegotiate()

        ntlm_negotiate_pkg = \
            proto._dcom.get_negotiate_ntlm_pkg(
                iid,
                ntlm_auth_negotiate,
                proto._auth_level)

        rpc_bind_ack: RpcBindAck = \
            await proto.get_dcom_response(ntlm_negotiate_pkg)

        proto._auth_type = rpc_bind_ack.auth.auth_type
        proto._auth_level = rpc_bind_ack.auth.auth_level

        ntlm_auth_challenge = NTLMAuthChallenge(rpc_bind_ack.auth.auth_value)
        proto._flags = flags = ntlm_auth_negotiate.get_negotiate_flags()

        ntlm_auth_authenticate = NTLMAuthAuthenticate(flags)

        session_base_key = ntlm_auth_authenticate.set_credentials(
            self._username,
            self._password,
            ntlm_auth_challenge.target_info,
            ntlm_auth_challenge.server_challenge,
            domain_name=self._domain)

        assert flags == 3767042613

        if flags & NTLMSSP_NEGOTIATE_KEY_EXCH:
            exported_session_key = get_rangom_bytes(16)

            encr_random_session_key = encrypted_session_key(
                session_base_key, exported_session_key)

            ntlm_auth_authenticate.set_encr_random_session_key(
                encr_random_session_key)
        else:
            exported_session_key = session_base_key

        ntlm_auth_authenticate.set_workstation_name(self._host)

        if proto._auth_level in (
                RPC_C_AUTHN_LEVEL_CONNECT,
                RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,
                RPC_C_AUTHN_LEVEL_PKT_PRIVACY):
            if proto._auth_type == RPC_C_AUTHN_WINNT:
                if flags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:
                    client_signing_key = sign_key(
                        flags,
                        exported_session_key)
                    server_signing_key = sign_key(
                        flags,
                        exported_session_key,
                        client_mode=False)
                    client_sealing_key = seal_key(
                        flags,
                        exported_session_key)
                    server_sealing_key = seal_key(
                        flags,
                        exported_session_key,
                        client_mode=False)

                    client_sealing_handle = \
                        ARC4.new(client_sealing_key).encrypt
                    server_sealing_handle = \
                        ARC4.new(server_sealing_key).encrypt

                    proto._client_seal = seal_func(
                        client_signing_key,
                        client_sealing_handle)
                    proto._server_seal = seal_func(
                        server_signing_key,
                        server_sealing_handle)
                    proto._client_sign = sign_func(
                        client_signing_key,
                        client_sealing_handle)
                    proto._server_sign = sign_func(
                        server_signing_key,
                        server_sealing_handle)

        ntlm_auth_pkg = \
            proto._dcom.get_authenticate_ntlm_pkg(
                ntlm_auth_authenticate,
                proto._auth_level)

        proto.write(ntlm_auth_pkg)

    async def negotiate_ntlm(self) -> Protocol:
        proto = self._protocol
        proto._auth_level = RPC_C_AUTHN_LEVEL_PKT_PRIVACY

        await self._bind(IID_IRemoteSCMActivator, proto)

        remote_create_instance = RemoteCreateInstance(
            CLSID_IWbemLevel1Login,
            IID_IWbemLevel1Login)

        rci_pkg = remote_create_instance.get_data()

        request = RpcRequest(op_num=4)
        request.set_pdu_data(rci_pkg)

        request_pkg = request.seal_data(proto)

        rpc_response: RpcResponse = \
            await proto.get_dcom_response(request_pkg, RpcResponse.SIZE)

        message = rpc_response.get_message(proto)

        interface = RemoteCreateInstanceResponse(self._host, message)

        proto._interface = interface
        proto._iid = IID_IWbemLevel1Login

        host, port = interface.get_binding()

        try:
            conn = self._loop.create_connection(
                lambda: Protocol(loop=self._loop),
                host=self._host,
                port=port)

            _, proto = await asyncio.wait_for(
                conn,
                timeout=self._timeout)
        except Exception as e:
            logging.warning(
                f'failed to connect to {self._host} ({port}); '
                f'fallback to {host} ({port})...')

            conn = self._loop.create_connection(
                lambda: Protocol(loop=self._loop),
                host=host,
                port=port)

            _, proto = await asyncio.wait_for(
                conn,
                timeout=self._timeout)

        proto._auth_level = max(
            interface.scm_reply_info_data.authn_hint,
            RPC_C_AUTHN_LEVEL_PKT_INTEGRITY)

        await self._bind(IID_IWbemLevel1Login, proto)

        return proto

    async def login_ntlm(
            self,
            proto: Protocol,
            namespace: str = 'root/cimv2'):
        if not namespace.startswith('//'):
            namespace = '//./' + namespace
        ntlm_login = NTLMLogin(namespace)
        ntlm_login_pkg = ntlm_login.get_data()
        interface = self._protocol._interface

        request = RpcRequest(op_num=6, uuid_str=interface.get_ipid())

        request.set_pdu_data(ntlm_login_pkg)

        request_pkg = request.sign_data(proto)

        rpc_response: RpcResponse = \
            await proto.get_dcom_response(request_pkg, RpcResponse.SIZE)

        message = rpc_response.get_message(proto)

        ntlm_login_resp = NTLMLoginResponse(message)

        proto._interface = ntlm_login_resp
        proto._interface.scm_reply_info_data = interface.scm_reply_info_data
        self._iid = IID_IWbemServices
        self._namespace = namespace



import struct
from typing import Union, Optional, List, TYPE_CHECKING
from .qualifier_set import QualifierSet
from ..cim_type import CimType
from .encoded_value import EncodedValue
from ..exceptions import WbemFalse


if TYPE_CHECKING:
    from ..protocol import Protocol
    from ..connection import Connection
    from ..ndr.next_response import NextResponse


class PropertyInfo:

    FMT = '<LHLL'
    FMT_SZ = struct.calcsize(FMT)

    type: int
    order: int

    def __init__(self, heap: bytes, entry: int) -> 'PropertyInfo':
        (
            self.type,
            self.order,
            value_table_offset,  # unused
            class_of_origin,  # unused
        ) = struct.unpack_from(self.FMT, heap, offset=entry)

        # see [MS-WMIO]: 2.2.32 Inherited
        # inherited = bool(self.type & CimType.CIM_INHERITED_FLAG)
        self.value = None
        self.null_default: Optional[bool] = None
        self.inherited_default: Optional[bool] = None

        # don't need the qualifiers, therefore make it optional using
        # set_qualifiers(..) and only store the offset
        self.qualifier_set: Union[QualifierSet, int] = entry + self.FMT_SZ

    def _set_qualifiers(self, heap: bytes):
        """Set qualifiers.

        The qualifiers contain info such as the CIMTYPE but they are not
        nesserary required. Therefore we do not unpack the qualifiers by
        default.
        """
        offset = self.qualifier_set
        self.qualifier_set, _ = QualifierSet.from_data(heap, offset)
        self.qualifier_set.load(heap)

    def _set_value(self, entry: int, heap: bytes):
        """Set value.

        This method can be called using different heap. From the parent class
        and from the actual class.
        """
        self.value = EncodedValue.get_value(self.type, entry, heap)

    def _set_type_default(self):
        p_type = self.type & (
            ~(CimType.CIM_ARRAY_FLAG | CimType.CIM_INHERITED_FLAG))

        if self.type & CimType.CIM_ARRAY_FLAG:
            self.value = []
        elif p_type == CimType.CIM_TYPE_BOOLEAN:
            self.value = False
        elif p_type == CimType.CIM_TYPE_OBJECT:
            self.value = None
        elif p_type in (
                CimType.CIM_TYPE_STRING,
                CimType.CIM_TYPE_DATETIME,
                CimType.CIM_TYPE_REFERENCE):
            self.value = ''
        else:
            self.value = 0

    def get_cim_type_name(self) -> str:
        """Return the name of the CIMTYPE.

        Note that in case of an array, the type of the array items will be
        returned. Thus, an array of strings would return `string`.
        """
        return CimType.get_cim_type_name(self.type)

    def is_array(self):
        """Return if this property is an array of not."""
        return self.type & CimType.CIM_ARRAY_FLAG

    def is_reference(self):
        """Return if this property is a reference."""
        p_type = self.type & ~CimType.CIM_INHERITED_FLAG
        return p_type == CimType.CIM_TYPE_REFERENCE

    def is_array_reference(self):
        """Return if this property is an array with references."""
        p_type = self.type & ~CimType.CIM_INHERITED_FLAG
        return p_type == CimType.CIM_ARRAY_REFERENCE

    def get_type(self) -> type:
        """Return the Python type for this property. Note that in case of an
        array type `list` will be returned. Use .get_cim_type() to get the
        type of items in a potential `list`.
        """
        if self.is_array():
            return list
        return CimType.get_cim_type_pytype(self.type)

    def get_cim_type(self) -> type:
        """Return the Python type for this property. In case of a `list`, the
        type for the items in the list will ne returned. For example, a list
        of strings would return type `str`. Use .get_type() if you want type
        list as a return type for a list of items.
        """
        return CimType.get_cim_type_pytype(self.type)

    async def _get_reference(
            self,
            value: str,
            conn: 'Connection',
            service: 'Protocol',
            filter_props: Optional[List[str]] = None) -> 'NextResponse':
        from ..query import Query

        value = value[value.find(':') + 1:]

        wmiclass, value = value.split('.', 1)

        # TODO: this is not perfect as a comma could technically exist in a
        # string so we should ignore a comma inside a string.
        constraints = ' AND '.join(value.split(','))

        props = ', '.join(filter_props) if filter_props else '*'

        query = Query(f'SELECT {props} FROM {wmiclass} WHERE {constraints}')

        await query.start(conn, service)
        res = await query.next()

        return res

    async def get_reference(
            self,
            conn: 'Connection',
            service: 'Protocol',
            filter_props: Optional[List[str]] = None) -> 'NextResponse':
        """This method raises WbemFalse if the reference does not exist."""
        res = await self._get_reference(
            self.value,
            conn,
            service,
            filter_props)
        return res

    async def get_array_references(
            self,
            conn: 'Connection',
            service: 'Protocol',
            filter_props: Optional[List[str]] = None,
            missing_as_none: bool = False) -> List['NextResponse']:
        arr = []
        for reference in self.value:
            try:
                res = await self._get_reference(
                    reference,
                    conn,
                    service,
                    filter_props)
            except WbemFalse:
                if missing_as_none:
                    res = None
                else:
                    raise
            arr.append(res)

        return arr



import struct
from ..tools import pad8


class NdrCommon:

    COMMON_FMT = '<BBHL'
    COMMON_SZ = struct.calcsize(COMMON_FMT)

    PRIVATE_FMT = '<LL'
    PRIVATE_SZ = struct.calcsize(PRIVATE_FMT)

    @classmethod
    def get_common(cls):
        version = 1
        endianness = 0x10
        common_header_len = cls.COMMON_SZ
        filter = 0xcccccccc

        return struct.pack(
            cls.COMMON_FMT,
            version,
            endianness,
            common_header_len,
            filter)

    @classmethod
    def get_private(cls, buffer: bytes):
        size = len(buffer)
        filter = 0xcccccccc
        data = struct.pack(
            cls.PRIVATE_FMT,
            size,
            filter)
        # padding is actually over the total, but since common + private = 16
        # this value is equal to the padding over the buffer
        return data, pad8(size) * b'\xFA'

    def read_common(self, data: bytes, offset: int):
        (
            self.version,
            self.endianness,
            self.common_header_len,
            filter
        ) = struct.unpack_from(self.COMMON_FMT, data, offset=offset)

    def read_private(self, data: bytes, offset: int):
        (
            self.size,
            filter,
        ) = struct.unpack_from(self.PRIVATE_FMT, data, offset=offset)
